#ifndef __IO_BUTTON_BUTTON__
#define __IO_BUTTON_BUTTON__

#include "Arduino.h"
#include "IO.h"

namespace io {
namespace button {

using io::Input;
using io::IO;

class Button;

enum State {
  PRESSED,
  RELEASED,
};

typedef struct {
public:
  Button& source;
  State state;
  unsigned long delta;
  unsigned long timer;
  unsigned short clicksCounter;
} Event;

void noop(const Event&);
typedef void (*OnPress)(const Event&);
typedef void (*OnRelease)(const Event&);
typedef void (*OnHold)(const Event&);

class Button : public Input<bool> {
public:
  class Listener;

  bool read() const;
  void setListener(Listener* listener);
  void unsetListener();
  void tick(unsigned long delta);

  void fireOnPress(unsigned long delta);
  void fireOnRelease(unsigned long delta);
  void fireOnHold(unsigned long delta);


  unsigned int getMultipleClickThreshold() const;
  void setMultipleClickThreshold(unsigned int);
  unsigned int getDebounceDelay() const;
  void setDebounceDelay(unsigned int);

  class Listener {
  public:
    virtual void onPress(const Event& event);
    virtual void onRelease(const Event& event);
    virtual void onHold(const Event& event);
  };

  template <typename OnPress, typename OnRelease, typename OnHold>
  class ListenerAdapter : public Listener {
  public:
    ListenerAdapter(OnPress onPress, OnRelease onRelease, OnHold onHold) : _onPress(onPress), _onRelease(onRelease), _onHold(onHold) {}

    virtual void onPress(const Event& event) { this->_onPress(event); }

    virtual void onRelease(const Event& event) { this->_onRelease(event); }

    virtual void onHold(const Event& event) { this->_onHold(event); }

  private:
    OnPress _onPress;
    OnRelease _onRelease;
    OnHold _onHold;
  };

  template <typename OnPress, typename OnRelease, typename OnHold>
  static ListenerAdapter<OnPress, OnRelease, OnHold> createListener(OnPress onPress, OnRelease onRelease, OnHold onHold) {
    return ListenerAdapter<OnPress, OnRelease, OnHold>(onPress, onRelease, onHold);
  }

  template <typename OnPress>
  static ListenerAdapter<OnPress, OnRelease, OnHold> onPress(OnPress onPress) {
    return ListenerAdapter<OnPress, OnRelease, OnHold>(onPress, noop, noop);
  }

  template <typename OnRelease>
  static ListenerAdapter<OnPress, OnRelease, OnHold> onRelease(OnRelease onRelease) {
    return ListenerAdapter<OnPress, OnRelease, OnHold>(noop, onRelease, noop);
  }

  template <typename OnHold>
  static ListenerAdapter<OnPress, OnRelease, OnHold> onHold(OnHold onHold) {
    return ListenerAdapter<OnPress, OnRelease, OnHold>(noop, noop, onHold);
  }

  static Button activeLowButton(Input<bool>* input) {
    return Button(input, true);
  }

  static Button activeHighButton(Input<bool>* input) {
    return Button(input, false);
  }

  Button(Input<bool>* input, bool reversed = false);
  virtual ~Button();

private:
  Input<bool>* input;
  Listener* listener;
  State state = State::RELEASED;
  bool reversed = false;
  unsigned short clickCounter = 0;
  unsigned long debounceTimer = 0;
  unsigned long lastPressTimer = 0;
  unsigned long lastReleaseTimer = 0;
  unsigned int debounceDelay = 20; // thinking about ms here
  unsigned int multipleClickThreshold = 250; // thinking about ms here
  bool lastValue;
  unsigned char pin;
};

} // namespace button

} // namespace io

#endif