#ifndef __IO_H__
#define __IO_H__

#include <Arduino.h>

namespace io {

template<typename V>
class Output {
public:
    virtual ~Output(){};
    virtual void write(V value) = 0;
};


template<typename V>
class WriteInAddressOutput: public Output<V> {
protected:
  V* address;
public:
  WriteInAddressOutput(V* address): address(address) {

  }
  
  void write(V value) {
    *(this->address) = value;
  }
};


template<typename V>
class Input {
public:
    virtual ~Input(){};
    virtual V read() const = 0;
};


template<typename V>
class ReadFromAddressInput: public Input<V> {
protected:
  const V* address;
public:
  ReadFromAddressInput(const V* address): address(address) {

  }

  V read() const {
    return *(this->address);
  }
};

template<typename V, typename F>
class OutputAdapter : public Output<V> {
protected:
    F fn;
public:
    OutputAdapter(F fn) : fn(fn) {}

    virtual void write(V value) {
        this->fn(value);
    }
};


template<typename V, typename F>
class InputAdapter : public Input<V> {
protected:
    F fn;
public:
    InputAdapter(F fn) : fn(fn) {}

    virtual V read() const {
        return this->fn();
    }
};

class BaseIO {
public:
    BaseIO() {};

    template<typename V, typename F>
    io::OutputAdapter<V, F> output(F fn) {
        return io::OutputAdapter<V, F>(fn);
    }

    template<typename V, typename F>
    io::InputAdapter<V, F> input(F fn) {
        return io::InputAdapter<V, F>(fn);
    }

    template<typename V>
    io::ReadFromAddressInput<V> inputFromAddress(const V* address) {
        return io::ReadFromAddressInput<V>(address);
    }

    template<typename V>
    io::WriteInAddressOutput<V>  outputFromAddress(V* address) {
      return io::WriteInAddressOutput<V>(address);
    }

};

}


#if defined(__LINUX_ARDUINO__) || defined(__AVR__)
#define USE_ARDUINO_IO
#include "io/ArduinoIO.h"
#endif

#if defined(ESP32) || defined(ESP_PLATFORM)
#define USE_ESP32_IO
#include "io/ESP32IO.h"
#endif

#endif