#include "io/button.h"
#include "Arduino.h"
#include <cmath>
#include <gtest/gtest.h>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>

using io::IO;
using io::button::Button;
using io::button::Event;

void debounceTick(Button& button) {
  unsigned int debounceDelay = button.getDebounceDelay();
  for (int i = 0; i < debounceDelay; i += 1) {
    button.tick(1);
  }
}


TEST(IO, DigialInputButton) {
  auto pin1 = IO.digitalInput(1);
  Button button(&pin1);
  for (int i = 0; i < 100; i++) {
    debounceTick(button);
  }
}

TEST(IO, AddressInputButton) {
  bool value = false;
  auto pin1 = IO.inputFromAddress(&value);
  Button button(&pin1);
  ASSERT_EQ(value, button.read());
}


TEST(IO, ButtonDebounce) {
  bool value = false;
  int pressCounter = 0;
  auto listener = Button::onPress([&pressCounter](const Event& event) mutable {
    ASSERT_TRUE(false);
  });
  auto pin1 = IO.inputFromAddress(&value);
  Button button(&pin1);
  button.setListener(&listener);
  for(int i=0;i<100;i++){
    button.tick(1);
    value = !value;
  }
}


TEST(IO, ButtonListenerOnPress) {
  bool value = false;
  int pressCounter = 0;
  auto listener = Button::onPress([&pressCounter](const Event& event) mutable {
    pressCounter += 1;
  });
  auto pin1 = IO.inputFromAddress(&value);
  Button button(&pin1);
  button.setListener(&listener);
  debounceTick(button);
  ASSERT_EQ(pressCounter, 0);
  value = true;
  debounceTick(button);
  ASSERT_EQ(pressCounter, 1);
}

TEST(IO, ButtonListenerOnRelease) {
  bool value = false;
  int releaseCounter = 0;
  auto listener = Button::onRelease([&releaseCounter](const Event& event) mutable {
    releaseCounter += 1;
  });
  auto pin1 = IO.inputFromAddress(&value);
  Button button(&pin1);
  button.setListener(&listener);
  debounceTick(button);
  ASSERT_EQ(releaseCounter, 0);
  value = true;
  debounceTick(button);
  ASSERT_EQ(releaseCounter, 0);
  value = false;
  debounceTick(button);
  ASSERT_EQ(releaseCounter, 1);
}

TEST(IO, ButtonListenerOnHold) {
  bool value = false;
  int holdCounter = 0;
  auto listener = Button::onHold([&holdCounter](const Event& event) mutable {
    holdCounter += 1;
  });
  auto pin1 = IO.inputFromAddress(&value);
  Button button(&pin1);
  button.setListener(&listener);
  debounceTick(button);
  ASSERT_EQ(holdCounter, 0);
  value = true;
  debounceTick(button);
  ASSERT_EQ(holdCounter, 0);
  value = true;
  debounceTick(button);
  ASSERT_GT(holdCounter, 0);
  value = false;
  debounceTick(button);
  ASSERT_GT(holdCounter, 0);
}

TEST(IO, ButtonPressSingleClick) {
  bool value = false;
  auto listener = Button::onPress([](const Event& event) {
    ASSERT_EQ(event.clicksCounter, 1);
  });
  auto pin1 = IO.inputFromAddress(&value);
  Button button(&pin1);
  debounceTick(button);
  button.setListener(&listener);
  value = true;
  debounceTick(button);
  value = false;
  debounceTick(button);
}

TEST(IO, ButtonPressDoubleClick) {
  bool value = false;
  auto listener = Button::onPress([](const Event& event) {
    ASSERT_EQ(event.clicksCounter, 2);
  });
  auto pin1 = IO.inputFromAddress(&value);
  Button button(&pin1);
  debounceTick(button);
  value = true;
  debounceTick(button);
  value = false;
  debounceTick(button);
  value = true;
  button.setListener(&listener);
  debounceTick(button);
}

TEST(IO, ButtonReleaseSingleClick) {
  bool value = false;
  auto listener = Button::onRelease([](const Event& event) {
    ASSERT_EQ(event.clicksCounter, 1);
  });
  auto pin1 = IO.inputFromAddress(&value);
  Button button(&pin1);
  debounceTick(button);
  button.setListener(&listener);
  value = true;
  debounceTick(button);
  value = false;
  debounceTick(button);
}

TEST(IO, ButtonReleaseDoubleClick) {
  bool value = false;
  auto listener = Button::onRelease([](const Event& event) {
    ASSERT_EQ(event.clicksCounter, 2);
  });
  auto pin1 = IO.inputFromAddress(&value);
  Button button(&pin1);
  debounceTick(button);
  value = true;
  debounceTick(button);
  value = false;
  debounceTick(button);
  value = true;
  button.setListener(&listener);
  debounceTick(button);
  value = false;
  debounceTick(button);
}

int main(int argc, char** argv) {
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}